#!/usr/bin/env python3

"""
Query robot base pose and speed from pybullet and publish to /odom topic
This component does not add any noise to it
"""

import rospy, tf
from nav_msgs.msg import Odometry

class simpleOdometry:
    def __init__(self, pybullet, robot, **kargs):
        # get "import pybullet as pb" and store in self.pb
        self.pb = pybullet
        # get robot from parent class
        self.robot = robot
        # register this node as a /odom publisher
        self.pub_odometry = rospy.Publisher('odom', Odometry, queue_size=1)
        # save some overhead by setting some information only once
        self.odom_msg = Odometry()
        self.odom_msg.header.frame_id = rospy.get_param('~odom_frame', 'odom')
        self.odom_msg.child_frame_id = rospy.get_param('~robot_base_frame', 'base_link')
        self.br = tf.TransformBroadcaster()

    def execute(self):
        """this function gets called from pybullet ros main update loop"""
        # set msg timestamp based on current time
        self.odom_msg.header.stamp = rospy.Time.now()
        # query base pose from pybullet and store in odom msg
        position, orientation = self.pb.getBasePositionAndOrientation(self.robot)
        [self.odom_msg.pose.pose.position.x,\
         self.odom_msg.pose.pose.position.y,\
         self.odom_msg.pose.pose.position.z] = position
        [self.odom_msg.pose.pose.orientation.x,\
         self.odom_msg.pose.pose.orientation.y,\
         self.odom_msg.pose.pose.orientation.z,\
        self.odom_msg.pose.pose.orientation.w] = orientation
        # query base velocity from pybullet and store it in msg
        [self.odom_msg.twist.twist.linear.x,\
         self.odom_msg.twist.twist.linear.y,\
         self.odom_msg.twist.twist.linear.z],\
        [self.odom_msg.twist.twist.angular.x,\
         self.odom_msg.twist.twist.angular.y,\
         self.odom_msg.twist.twist.angular.z] = self.pb.getBaseVelocity(self.robot)
        self.pub_odometry.publish(self.odom_msg)
        # tf broadcast (odom to base_link)
        self.br.sendTransform(position, orientation, rospy.Time.now(), \
            self.odom_msg.child_frame_id, self.odom_msg.header.frame_id)



#!/usr/bin/env python3

"""
RGBD camera sensor simulation for pybullet_ros base on pybullet.getCameraImage()
"""

import math
import numpy as np

import rospy
from cv_bridge import CvBridge
from sensor_msgs.msg import Image, CameraInfo
import os
import yaml

class RGBDCamera:
    def __init__(self, pybullet, robot, **kargs):
        # get "import pybullet as pb" and store in self.pb
        self.pb = pybullet
        # get robot from parent class
        self.robot = robot
        # create image msg placeholder for publication
        self.image_msg = Image()
        # get RGBD camera parameters from ROS param server
        self.image_msg.width = rospy.get_param('~rgbd_camera/resolution/width', 640)
        self.image_msg.height = rospy.get_param('~rgbd_camera/resolution/height', 480)
        assert(self.image_msg.width > 5)
        assert(self.image_msg.height > 5)
        cam_frame_id = rospy.get_param('~rgbd_camera/frame_id', "kinect_link")
        if not cam_frame_id:
            rospy.logerr('Required parameter rgbd_camera/frame_id not set, will exit now...')
            rospy.signal_shutdown('Required parameter rgbd_camera/frame_id not set')
            return
        # get pybullet camera link id from its name
        link_names_to_ids_dic = kargs['link_ids']
        if not cam_frame_id in link_names_to_ids_dic:
            rospy.logerr('Camera reference frame "{}" not found in URDF model'.format(cam_frame_id))
            rospy.logwarn('Available frames are: {}'.format(link_names_to_ids_dic))
            rospy.signal_shutdown('required param rgbd_camera/frame_id not set properly')
            return
        self.pb_camera_link_id = link_names_to_ids_dic[cam_frame_id]
        self.image_msg.header.frame_id = cam_frame_id
        # create publisher
        self.pub_image = rospy.Publisher('kinect/rgb/image_raw', Image, queue_size=10)
        self.image_msg.encoding = rospy.get_param('~rgbd_camera/resolution/encoding', 'rgb8')
        self.image_msg.is_bigendian = rospy.get_param('~rgbd_camera/resolution/encoding', 0)
        self.image_msg.step = rospy.get_param('~rgbd_camera/resolution/encoding', 1920)
        # projection matrix
        self.hfov = rospy.get_param('~rgbd_camera/hfov', 56.3)
        self.vfov = rospy.get_param('~rgbd_camera/vfov', 43.7)
        self.near_plane = rospy.get_param('~rgbd_camera/near_plane', 0.4)
        self.far_plane = rospy.get_param('~rgbd_camera/far_plane', 8)
        self.projection_matrix = self.compute_projection_matrix()
        # use cv_bridge ros to convert cv matrix to ros format
        self.image_bridge = CvBridge()
        # variable used to run this plugin at a lower frequency, HACK
        self.count = 0
        # publishing depth/camera_info topic
        depth_file = os.path.join(os.path.dirname(__file__), "depth_msg.yaml")
        with open(depth_file, "r") as depth_file_handle:
            depth_data = yaml.load(depth_file_handle)
        self.depth_camera_info_msg = CameraInfo()
        self.depth_camera_info_msg.width = depth_data["image_width"]
        self.depth_camera_info_msg.height = depth_data["image_height"]
        self.depth_camera_info_msg.K = depth_data["camera_matrix"]["data"]
        self.depth_camera_info_msg.D = depth_data["distortion_coefficients"]["data"]
        self.depth_camera_info_msg.R = depth_data["rectification_matrix"]["data"]
        self.depth_camera_info_msg.P = depth_data["projection_matrix"]["data"]
        self.pub_depth_msg = rospy.Publisher('kinect/depth/camera_info', CameraInfo, queue_size=10)
        # publishing rgb/camera_info topic
        rgb_file = os.path.join(os.path.dirname(__file__), "rgb_msg.yaml")
        with open(rgb_file, "r") as depth_file_handle:
            rgb_data = yaml.load(depth_file_handle)
        self.rgb_camera_info_msg = CameraInfo()
        self.rgb_camera_info_msg.width = rgb_data["image_width"]
        self.rgb_camera_info_msg.height = rgb_data["image_height"]
        self.rgb_camera_info_msg.K = rgb_data["camera_matrix"]["data"]
        self.rgb_camera_info_msg.D = rgb_data["distortion_coefficients"]["data"]
        self.rgb_camera_info_msg.R = rgb_data["rectification_matrix"]["data"]
        self.rgb_camera_info_msg.P = rgb_data["projection_matrix"]["data"]
        self.pub_rgb_msg = rospy.Publisher('kinect/rgb/camera_info', CameraInfo, queue_size=10)

    def compute_projection_matrix(self):
        return self.pb.computeProjectionMatrix(
                    left=-math.tan(math.pi * self.hfov / 360.0) * self.near_plane,
                    right=math.tan(math.pi * self.hfov / 360.0) * self.near_plane,
                    bottom=-math.tan(math.pi * self.vfov / 360.0) * self.near_plane,
                    top=math.tan(math.pi * self.vfov / 360.0) * self.near_plane,
                    nearVal=self.near_plane,
                    farVal=self.far_plane)

    def extract_frame(self, camera_image):
        bgr_image = np.zeros((self.image_msg.height, self.image_msg.width, 3))

        camera_image = np.reshape(camera_image[2], (camera_image[1], camera_image[0], 4))

        bgr_image[:, :, 2] =\
            (1 - camera_image[:, :, 3]) * camera_image[:, :, 2] +\
            camera_image[:, :, 3] * camera_image[:, :, 2]

        bgr_image[:, :, 1] =\
            (1 - camera_image[:, :, 3]) * camera_image[:, :, 1] +\
            camera_image[:, :, 3] * camera_image[:, :, 1]

        bgr_image[:, :, 0] =\
            (1 - camera_image[:, :, 3]) * camera_image[:, :, 0] +\
            camera_image[:, :, 3] * camera_image[:, :, 0]

        # return frame
        return bgr_image.astype(np.uint8)

    def compute_camera_target(self, camera_position, camera_orientation):
        """
        camera target is a point 5m in front of the robot camera
        This method is used to tranform it to the world reference frame
        NOTE: this method uses pybullet functions and not tf
        """
        target_point = [5.0, 0, 0] # expressed w.r.t camera reference frame
        camera_position = [camera_position[0], camera_position[1], camera_position[2]]
        rm = self.pb.getMatrixFromQuaternion(camera_orientation)
        rotation_matrix = [[rm[0], rm[1], rm[2]],[rm[3], rm[4], rm[5]],[rm[6], rm[7], rm[8]]]
        return np.dot(rotation_matrix, target_point) + camera_position

    def execute(self):
        """this function gets called from pybullet ros main update loop"""
        # run at lower frequency, camera computations are expensive
        # self.count += 1
        # if self.count < 100:
        #     return
        # self.count = 0 # reset count
        # get camera pose
        cam_state = self.pb.getLinkState(self.robot, self.pb_camera_link_id)
        # target is a point 5m ahead of the robot camera expressed w.r.t world reference frame
        target = self.compute_camera_target(cam_state[0], cam_state[1])
        view_matrix = self.pb.computeViewMatrix(cam_state[0], target, [0, 0, 1])
        # get camera image from pybullet
        pybullet_cam_resp = self.pb.getCameraImage(self.image_msg.width,
                                                   self.image_msg.height,
                                                   view_matrix,
                                                   self.projection_matrix,
                                                   renderer=self.pb.ER_BULLET_HARDWARE_OPENGL,
                                                   flags=self.pb.ER_NO_SEGMENTATION_MASK)
        # frame extraction function from qibullet
        frame = self.extract_frame(pybullet_cam_resp)
        # fill pixel data array
        self.image_msg.data = self.image_bridge.cv2_to_imgmsg(frame).data
        # update msg time stamp
        self.image_msg.header.stamp = rospy.Time.now()
        # publish camera image to ROS network
        self.pub_image.publish(self.image_msg)
        self.pub_rgb_msg.publish(self.rgb_camera_info_msg)
        self.pub_depth_msg.publish(self.depth_camera_info_msg)



#!/usr/bin/env python3

"""
Imagine you have to run multiple functions within a deadline.
Running them sequentially will be the easiest thing to do, but if one of them behaves badly
and takes too much time, we are in trouble...

This code, instead of runing functions sequentially, we run them all on separate threads in parallel.
This allows us to consistently call (in time) the functions that meet their deadline, while
ignoring those functions that miss the cycle rate, yet still call them at their maximum rate.

Example:

             plugin1       plugin2         plugin3
start    ----------------------------------------------
              start        start           start
                
              finish
                
                                           finish




loop rate ----------------------------------------------  (deadline)
              start again  start again
                                           finish
                                           start again

                                        .
                                        .
                                        . etc...

  In this example plugin1 and 2 meet their deadline (and therefore are "in time"),
  while plugin3 takes more time than the desired call frequency (misses the loop rate), yet
  it is still being called as fast as it can make its computation.

See function_exec_manager_example.py for a runing example on how to use this class
"""

from threading import Thread
import time, sys, copy

class FuncExecManager:
    """
    helper class to keep track of synchronous multiple parallel execution of functions with deadlines
    """
    def __init__(self, list_of_objects, stop_condition, exec_after_each_loop, pause_execution, log_info=print,
                 log_warn=print, log_debug=print, function_name='function'):
        # deadline for functions to finish their process
        self.loop_rate = 0.25 # every 4 secs
        # keep track of functions which execution time is "below" the deadline (set by loop rate)
        self.on_time_functions = []
        # keep track of functions which execution time is "above" the deadline (set by loop rate)
        self.late_functions = []
        self.late_threads = []
        # flag to know when a cycle is finished
        self.is_loop_finished = False
        # to differentiate between each cycle (and keep track of functions that finish on time)
        self.cycle_unique_id = None
        # save in member variable the received list of objects
        self.list_of_objects = list_of_objects
        # the function that will be called to check if algorithm should stop or continue
        self.stop_condition = stop_condition
        # this function will be called after each loop
        self.exec_after_each_loop = exec_after_each_loop
        # offer the user the possibility to pause the execution
        self.pause_execution = pause_execution
        # configure loggers
        log_info(f'Started synchronous {function_name} execution manager')
        self.log_warn = log_warn
        self.log_debug = log_debug

    def time_control(self, cycle_unique_id, obj):
        class_name = str(obj.__class__)
        # make backup of wall time
        start_time = time.time()
        obj.execute()
        # compare id to check if we are still on time
        if self.cycle_unique_id == cycle_unique_id:
            self.log_debug(f'finished {class_name} in time')
            self.on_time_functions.append(obj)
        else:
            end_time = time.time()
            self.log_warn(f'{class_name}: Missed loop rate, took {str(round(end_time - start_time - (1.0 / self.loop_rate), 2))} sec longer than expected')
            self.late_functions.append(obj)

    def loop_thread(self):
        """
        we create an additional thread to monitor the deadline
        """
        # convert frequency to time
        time.sleep(1.0 / self.loop_rate)
        self.log_debug('==== loop rate! ====')
        # raise flag to indicate that one loop is complete
        self.is_loop_finished = True

    def start_synchronous_execution(self, loop_rate=0.25):
        # allow user to modify the frequency at which the functions will be called
        self.loop_rate = loop_rate
        # at first all functions are on time
        self.on_time_functions = self.list_of_objects
        self.late_functions = []
        # init cycle_unique_id
        cycle_unique_id = 0
        if cycle_unique_id > sys.maxsize:
            cycle_unique_id = 0
        # run x iterations until user wants to stop (you might want to pass here a ctrl + c detection)
        while not self.stop_condition():
            # initialize flag
            self.is_loop_finished = False
            # jump to the next cycle id (because previous loop is finished)
            cycle_unique_id = cycle_unique_id + 1
            self.cycle_unique_id = cycle_unique_id
            # start "on time" functions
            thread_list = []
            for func in self.on_time_functions:
                thread_list.append(Thread(target=self.time_control, args=(cycle_unique_id, func,)))
            self.on_time_functions = []
            for t in thread_list:
                t.start()
            # run loop function on a separate thread
            Thread(target=self.loop_thread).start()
            # wait until loop finishes
            while not self.is_loop_finished:
                # if there is a late thread that finished, run it again right away
                if self.late_functions != []:
                    self.late_threads = []
                    for func in self.late_functions:
                        self.late_threads.append(Thread(target=self.time_control, args=(cycle_unique_id, func,)))
                        self.late_threads[-1].start()
                    self.late_functions = []
                # sleep to reduce computational load
                time.sleep(0.001)
                # check if user wants to pause execution
                while self.pause_execution():
                    time.sleep(0.1)
            # execute custom function after having finished the loop
            self.exec_after_each_loop()
        # wait for threads to finish
        for thread in self.late_threads:
            thread.join()



#!/usr/bin/env python3

"""
plugin that is loaded one time only at the beginning
It is meant to be for you to upload your environment
"""

import rospy
import os

ENV_PATH = '/home/saun/My_DQN_MB_model/src/turtlebot3_simulations/turtlebot3_gazebo/models/training_maze/env.urdf'
class Environment:
    def __init__(self, pybullet, **kargs):
        # get "import pybullet as pb" and store in self.pb
        self.pb = pybullet
        # enable soft body simulation if needed
        if rospy.get_param('~use_deformable_world', False):
            rospy.loginfo('Using deformable world (soft body simulation)')
            self.pb.resetSimulation(self.pb.RESET_USE_DEFORMABLE_WORLD)

    def load_environment(self):
        """
        set gravity, ground plane and load URDF or SDF models as required
        """
        # set gravity
        self.pb.setGravity(0, 0, -9.81)
        # set floor
        plane_pos = rospy.get_param("~plane_pos", [0, 0, 0])
        self.pb.loadURDF('plane.urdf', basePosition=plane_pos)
        self.env = self.pb.loadURDF(ENV_PATH)
        org_env_orientation = self.pb.getQuaternionFromEuler([0,0,0])
        self.pb.resetBasePositionAndOrientation(self.env, [0,0,0],org_env_orientation)
        self.load_environment_via_code()

    def load_environment_via_code(self):
        """
        This method provides the possibility for the user to define an environment via python code
        example:
        self.pb.loadURDF(...)
        self.pb.loadSDF(...)
        self.pb.loadSoftBody(...)
        self.pb.setTimeStep(0.001)
        self.pb.setPhysicsEngineParameter(sparseSdfVoxelSize=0.25) # ? related to soft bodies
        etc...
        NOTE: is not advised to directly write code below, instead make new class and inherit from this one
              see example: environment_template.py
        """
        pass



import setuptools
from pathlib import Path

setuptools.setup(
    name='haibot_rosgymbullet',
    version='0.4.3',
    description="DiffBot Ros-Gym-Bullet Environment",
    long_description=Path("README.md").read_text(),
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(include="haibot_rosgymbullet*"),
    install_requires=['gym', 'pybullet', 'numpy', 'matplotlib']
)